/*********************************************************************
 *
 *                Microchip USB C18 Firmware Version 1.0
 *
 *********************************************************************
 * FileName:        user.c
 * Dependencies:    See INCLUDES section below
 * Processor:       PIC18
 * Compiler:        C18 2.30.01+
 * Company:         Microchip Technology, Inc.
 *
 * Software License Agreement
 *
 * The software supplied herewith by Microchip Technology Incorporated
 * (the Company) for its PICmicro Microcontroller is intended and
 * supplied to you, the Companys customer, for use solely and
 * exclusively on Microchip PICmicro Microcontroller products. The
 * software is owned by the Company and/or its supplier, and is
 * protected under applicable copyright laws. All rights are reserved.
 * Any use in violation of the foregoing restrictions may subject the
 * user to criminal sanctions under applicable laws, as well as to
 * civil liability for the breach of the terms and conditions of this
 * license.
 *
 * THIS SOFTWARE IS PROVIDED IN AN AS IS CONDITION. NO WARRANTIES,
 * WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED
 * TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE COMPANY SHALL NOT,
 * IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL OR
 * CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
 *
 * Author               Date        Comment
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Rawin Rojvanit       11/19/04    Original.
 * Brian Schmalz		03/15/06	Added user code to impliment
 *									firmware version D v1.0 for UBW
 *									project. See www.greta.dhs.org/UBW
 * Brian Schmalz		05/04/06	Starting version 1.1, which will 
 * 									include several fixes. See website.
 * BPS					06/21/06	Starting v1.2 -
 * - Fixed problem with I packets (from T command) filling up TX buffer
 * 		and not letting any incoming commands be received. (strange)
 * - Adding several commands - Analog inputs being the biggest set.
 * - Also Byte read/Byte write (PEEK/POKE) anywhere in memory
 * - Individual pin I/O and direction
 * BPS					08/16/06	v1.3 - Fixed bug with USB startup
 ********************************************************************/

/** I N C L U D E S **********************************************************/
#include <p18cxxx.h>
#include <usart.h>
#include <stdio.h>
#include "system\typedefs.h"
#include "system\usb\usb.h"

#include "io_cfg.h"             // I/O pin mapping
#include "user\user.h"

/** V A R I A B L E S ********************************************************/
#define MAX_USB_BUF_SIZE 		64
#define MAX_WORKING_BUF_SIZE	128

#define kCHAR					0
#define kUCHAR					1
#define kINT					2
#define kUINT					3
#define kASCII_CHAR				4

#define kISR_FIFO_A_DEPTH		3
#define kISR_FIFO_D_DEPTH		3
#define kPR2_RELOAD				250		// For 1ms TMR2 tick

#define bitset(var,bitno) ((var) |= (1 << (bitno)))
#define bitclr(var,bitno) ((var) &= ~(1 << (bitno)))
#define bittst(var,bitno) (var& (1 << (bitno)))

#pragma udata
byte old_sw;

unsigned char input_buffer[MAX_USB_BUF_SIZE];
unsigned char output_buffer[MAX_USB_BUF_SIZE];
// Our error byte
char g_error;
// Version string
rom unsigned char g_version[]={"UBW FW D Version 1.3.0\n\r"};

// This big boy goes in a different bank
#pragma udata big_buf=0x200
unsigned char working_buf[MAX_WORKING_BUF_SIZE];
volatile unsigned int ISR_A_FIFO[12][kISR_FIFO_A_DEPTH];	// Stores the most recent analog conversions
volatile unsigned char ISR_D_FIFO[3][kISR_FIFO_D_DEPTH];	// FIFO of actual data

#pragma udata
// In and out pointers to our working USB data input buffer
unsigned char working_buf_in;
unsigned char working_buf_out;

// Rate variable - how fast does interrupt fire to capture inputs?
unsigned int time_between_updates;

volatile unsigned int ISR_D_RepeatRate;					// How many 1ms ticks between Digital updates
volatile unsigned char ISR_D_FIFO_in;					// In pointer
volatile unsigned char ISR_D_FIFO_out;					// Out pointer
volatile unsigned char ISR_D_FIFO_length;				// Current FIFO depth

volatile unsigned int ISR_A_RepeatRate;					// How many 1ms ticks between Analog updates
volatile unsigned char ISR_A_FIFO_in;					// In pointer
volatile unsigned char ISR_A_FIFO_out;					// Out pointer
volatile unsigned char ISR_A_FIFO_length;				// Current FIFO depth
volatile unsigned char AnalogEnable;					// Maximum ADC channel to convert

/** P R I V A T E  P R O T O T Y P E S ***************************************/
void add_to_buffer(unsigned char);
BOOL do_we_have_a_full_packet(unsigned char *);
void BlinkUSBStatus(void);
BOOL SwitchIsPressed(void);
void parse_packet(unsigned char);
signed short long extract_number(
	unsigned char * ptr, 
	unsigned char type
);
signed char extract_digit(
	unsigned char * ptr, 
	signed short long * acc, 
	unsigned char digits
);
void parse_R_packet(void);		// R for resetting UBW
void parse_C_packet(void);		// C for configuring I/O and analog pins
void parse_CX_packet(void); 	// CX For configuring serial port
void parse_O_packet(void);		// O for output digital to pins
void parse_I_packet(void);		// I for input digital from pins
void parse_V_packet(void);		// V for printing version
void parse_A_packet(void);		// A for requesting analog inputs
void parse_T_packet(void);		// T for setting up timed I/O (digital or analog)
void parse_PI_packet(void);		// PI for reading a single pin
void parse_PO_packet(void);		// PO for setting a single pin state
void parse_PD_packet(void);		// PD for setting a pin's direction
void parse_MR_packet(void);		// MR for Memory Read
void parse_MW_packet(void); 	// MW for Memory Write
void parse_TX_packet(void);		// TX for transmitting serial
void parse_RX_packet(void);		// RX for receiving serial
void parse_RC_packet(void);		// RC is for outputing RC servo pulses 

/** D E C L A R A T I O N S **************************************************/
#pragma code

#pragma interruptlow LowISR
void LowISR(void)
{
	static unsigned int D_tick_counter = 0;
	static unsigned int A_tick_counter = 0;
	static unsigned char A_cur_channel;
	
	// Do we have a Timer2 interrupt? (1ms rate)
	if (PIR1bits.TMR2IF)
	{
		// Clear the interrupt 
		PIR1bits.TMR2IF = 0;
		
		// See if it's time to fire off an I packet
		if (ISR_D_RepeatRate > 0)
		{
			D_tick_counter++;
			if (D_tick_counter >= ISR_D_RepeatRate)
			{
				D_tick_counter = 0;
				// Tell the main code to send an I packet
				if (ISR_D_FIFO_length < kISR_FIFO_D_DEPTH)
				{
					// And copy over our port values
					ISR_D_FIFO[0][ISR_D_FIFO_in] = PORTA;
					ISR_D_FIFO[1][ISR_D_FIFO_in] = PORTB;
					ISR_D_FIFO[2][ISR_D_FIFO_in] = PORTC;
					ISR_D_FIFO_in++;
					if (ISR_D_FIFO_in >= kISR_FIFO_D_DEPTH)
					{
						ISR_D_FIFO_in = 0;	
					}
					ISR_D_FIFO_length++;
				}
				else
				{
					// Stop the madness! Something is wrong, we're
					// not getting our packets out. So kill the 
					// timer.
					ISR_D_RepeatRate = 0;
				}
			}	
		}
		
		// See if it's time to fire off an A packet
		if ((ISR_A_RepeatRate > 0) && (AnalogEnable > 0))
		{
			A_tick_counter++;
			if (A_tick_counter >= ISR_A_RepeatRate)
			{
				A_tick_counter = 0;
				// Tell the main code to send an A packet
				if (ISR_A_FIFO_length < kISR_FIFO_A_DEPTH)
				{
					ISR_A_FIFO_in++;
					if (ISR_A_FIFO_in >= kISR_FIFO_A_DEPTH)
					{
						ISR_A_FIFO_in = 0;	
					}
					ISR_A_FIFO_length++;
				}
				else
				{
					// Stop the madness! Something is wrong, we're
					// not getting our packets out. So kill the A
					// packets.
					ISR_A_RepeatRate = 0;
				}
			}	
		}

		// See if it's time to start analog conversions
		if (AnalogEnable > 0)
		{
			// Set the channel to zero to start off with
			A_cur_channel = 0;
			ADCON0 = (A_cur_channel << 2) + 1;

			// Clear the interrupt
			PIR1bits.ADIF = 0;

			// And make sure to always use low priority.
			IPR1bits.ADIP = 0;

			// Set the interrupt enable
			PIE1bits.ADIE = 1;

			// Make sure it's on!
			ADCON0bits.ADON = 1;

			// And tell the A/D to GO!
			ADCON0bits.GO_DONE = 1;
		}
		
	}

	// Do we have an analog interrupt?
	if (PIR1bits.ADIF)
	{
		// Clear the interrupt
		PIR1bits.ADIF = 0;

		// Read out the value that we just converted, and store it.
		ISR_A_FIFO[A_cur_channel][ISR_A_FIFO_in] = 
			(unsigned int)ADRESL 
			| 
			((unsigned int)ADRESH << 8);
	
		// Incriment the channel and write the new one in 
		A_cur_channel++;
		if (A_cur_channel >= AnalogEnable)
		{
			// We're done, so just sit and wait
			// Turn off our interrupts though.
			PIE1bits.ADIE = 0;
		}
		else
		{
			// Update the channel number
			ADCON0 = (A_cur_channel << 2) + 1;
			// And start the next conversion
			ADCON0bits.GO_DONE = 1;
		}
	}
}


#pragma interrupt HighISR
void HighISR(void)
{
}

void UserInit(void)
{
	char i, j;

	// Make all of PORTA digital inputs
	LATA = 0x00;
	TRISA = 0xFF;
	// Turn all analog inputs into digital inputs
	ADCON1 = 0x0F;
	// Turn off the ADC
	ADCON0bits.ADON = 0;
	// Turn off our own idea of how many analog channels to convert
	AnalogEnable = 0;
	CMCON = 0x07;	// Comparators as digital inputs
	// Make all of PORTB inputs
	LATB = 0x00;
	TRISB = 0xFF;
	// Make all of PORTC inputs
	LATC = 0x00;
	TRISC = 0xFF;

	// Initalize LED I/Os to outputs
    mInitAllLEDs();
	// Initalize switch as an input
    mInitSwitch();
	// Save the current state of the switch
    old_sw = UserSW;

    // Inialize all of our pointers and stuff
    working_buf_in = 0;
    working_buf_out = 0;
    
	// Initalize all of the ISR FIFOs
    ISR_A_FIFO_out = 0;
    ISR_A_FIFO_in = 0;
    ISR_A_FIFO_length = 0;
    ISR_D_FIFO_out = 0;
    ISR_D_FIFO_in = 0;
    ISR_D_FIFO_length = 0;

	// Make sure that our timer stuff starts out disabled
	ISR_D_RepeatRate = 0;
	ISR_A_RepeatRate = 0;

    // Now init our registers
	// The prescaler will be at 16
    T2CONbits.T2CKPS1 = 1;
    T2CONbits.T2CKPS0 = 1;
    // We want the TMR2 post scaler to be a 3
    T2CONbits.T2OUTPS3 = 0;
    T2CONbits.T2OUTPS2 = 0;
    T2CONbits.T2OUTPS1 = 1;
    T2CONbits.T2OUTPS0 = 0;
	// Set our reload value
	PR2 = kPR2_RELOAD;

	// Set up the Analog to Digital converter
	// Clear out the FIFO data
	for (i = 0; i < 12; i++)
	{
		for (j = 0; j < kISR_FIFO_A_DEPTH; j++)
		{
			ISR_A_FIFO[i][j] = 0;
		}
	}	

    // Enable interrupt priorities
    RCONbits.IPEN = 1;
	T2CONbits.TMR2ON = 0;
    
    PIE1bits.TMR2IE = 1;
    IPR1bits.TMR2IP = 0;
    
    INTCONbits.GIEH = 1;
    INTCONbits.GIEL = 1;

	// Turn on the timer.
	T2CONbits.TMR2ON = 1;    

}//end UserInit

/******************************************************************************
 * Function:        void ProcessIO(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        In this function, we check for a new packet that just
 * 					arrived via USB. We do a few checks on the packet to see
 *					if it is worthy of us trying to interpret it. If it is,
 *					we go and call the proper function based upon the first
 *					character of the packet.
 *					NOTE: We need to see everything in one packet (i.e. we
 *					won't treat the USB data as a stream and try to find our
 *					start and end of packets within the stream. We just look 
 *					at the first character of each packet for a command and
 * 					check that there's a CR as the last character of the
 *					packet.
 *
 * Note:            None
 *****************************************************************************/
void ProcessIO(void)
{   
	char err = 1;
	char bytesIn = 0;
	static char LastPacketByte = 'A';   
	char bad = 0;
	char i = 0;
	unsigned char length;
	static unsigned int TryCount = 0;
	char snd_cntr;
	char pos;
	char channel;

	BlinkUSBStatus();

    // User Application USB tasks
    if((usb_device_state < CONFIGURED_STATE) || (UCONbits.SUSPND == 1))
	{	
		return;
	}

	// Check for a new I packet (from T command) ready to go out
	if (ISR_D_FIFO_length > 0 && mUSBUSARTIsTxTrfReady())
	{
		sprintf (
			(char *)output_buffer, 
			(rom char*)"I,%03i,%03i,%03i\n\r", 
			ISR_D_FIFO[0][ISR_D_FIFO_out],
			ISR_D_FIFO[1][ISR_D_FIFO_out],
			ISR_D_FIFO[2][ISR_D_FIFO_out]
		);
		ISR_D_FIFO_out++;
		if (ISR_D_FIFO_out == kISR_FIFO_D_DEPTH)
		{
			ISR_D_FIFO_out = 0;
		}
		ISR_D_FIFO_length--;
		putsUSBUSART(output_buffer);
	}			

	// Check for a new A packet (from T command) ready to go out
	if (ISR_A_FIFO_length > 0 && mUSBUSARTIsTxTrfReady())
	{
		pos = 0;

		// Put the beginning of the packet in place
		output_buffer[pos] = 'A';
		pos++;

		// Now add each analog value
		for (channel = 0; channel < AnalogEnable; channel++)
		{
			sprintf(
				 (char *)&output_buffer[pos]
				,(rom char *)",%04u" 
				,ISR_A_FIFO[channel][ISR_A_FIFO_out]
			);
			pos += 5;
		}

		// Add \n\r and terminating zero.
		output_buffer[pos] = 10;
		pos++;
		output_buffer[pos] = 13;
		pos++;
		output_buffer[pos] = 0x00;

		ISR_A_FIFO_out++;
		if (ISR_A_FIFO_out == kISR_FIFO_A_DEPTH)
		{
			ISR_A_FIFO_out = 0;
		}
		ISR_A_FIFO_length--;
		putsUSBUSART(output_buffer);
	}			

	// Pull in some new data if there is new data to pull in
	bytesIn = getsUSBUSART(input_buffer, MAX_USB_BUF_SIZE);
	
	g_error = 0;
	
	if (bytesIn > 0)
	{
		// We're going to add whatever we just got to our
		// buffer.
		add_to_buffer (bytesIn);
	}
	
	// Check to see if we have at least one full packet.
	if (do_we_have_a_full_packet (&length))
	{
		// This is where all of the hard work happens
		parse_packet (length);
		
		// Check to see if an error has been set.
		if (g_error > 0 && mUSBUSARTIsTxTrfReady())
		{
			output_buffer[0] = g_error;
			output_buffer[1] = 0x0D;
			output_buffer[2] = 0x0A;
			output_buffer[3] = 0x00;
			putsUSBUSART(output_buffer);
		}
	}
}

// Here, we need to add the last bunch of bytes to our working buffer.
// As we do, we strip out any combination of <CR><LF> and make them
// just a single <CR>.
void add_to_buffer(unsigned char length)
{
	unsigned char i;
	static BOOL in_cr = FALSE;
//	unsigned char orig_ptr = working_buf_in;

	// March through, adding each byte to the working buffer
	for (i=0; i < length; i++)
	{
		if (
			!in_cr 
			&& 
			(
				0x0D == input_buffer[i]
				||
				0x0A == input_buffer[i]
			)
		)
		{
			in_cr = TRUE;
			working_buf[working_buf_in] = 0x0D;
			working_buf_in++;
		}
		else if (
			input_buffer[i] != 0x0D
			&&
			input_buffer[i] != 0x0A
		)
		{
			working_buf[working_buf_in] = input_buffer[i];
			in_cr = FALSE;
			working_buf_in++;
		}
		if (MAX_WORKING_BUF_SIZE == working_buf_in)
		{
			working_buf_in = 0;
		}
		if (working_buf_in == working_buf_out)
		{
			break;
		}
	}


//	if (mUSBUSARTIsTxTrfReady())
//	{
//		sprintf( 
//			output_buffer, 
//			(rom char*)"OP=%d NP=%d LN=%d\r"
//			,orig_ptr
//			,working_buf_in
//			,length
//		);
//		putsUSBUSART(output_buffer);
//	}

}

// Scan through the front of the working buffer to see if we
// have a full packet (anything terminated by <CR>) and if so,
// return TRUE and put the length of the packet in 'length'.
// Otherwise return FALSE and set length to zero.
// We copy the packet over to input_buffer[].
BOOL do_we_have_a_full_packet(unsigned char * length)
{
	unsigned char i;
	unsigned char tst_char = working_buf_out;

	if (tst_char == working_buf_in)
	{
//		if (mUSBUSARTIsTxTrfReady())
//		{
//			sprintf( 
//				output_buffer, 
//				(rom char*)"TC=%d WBO=%d WBI=%d:\r"
//				,tst_char
//				,working_buf_out
//				,working_buf_in
//			);
//			putsUSBUSART(output_buffer);
//		}

		*length = 0;
		input_buffer[0] = 0x00;
		return (FALSE);
	}

	// We did not find our end, which is OK.
	for (i=0; i < MAX_WORKING_BUF_SIZE; i++)
	{
		// Copy over the next character of the packet
		input_buffer[i] = working_buf[tst_char];

		// Move the pointer in the working buffer up by one
		// wrapping around if necessary
		tst_char++;
		if (MAX_WORKING_BUF_SIZE == tst_char)
		{
			tst_char = 0;	
		}

		// Have we found our end?
		if (0x0D == input_buffer[i])
		{
			// Send back the length of this packet
			// (Add one because we're counting how many
			// bytes are in the packet, including the first
			// one)
			*length = i + 1;

//			if (mUSBUSARTIsTxTrfReady())
//			{
//				sprintf( 
//					output_buffer, 
//					(rom char*)"TC=%d WBO=%d WBI=%d LEN=%d:\r"
//					,tst_char
//					,working_buf_out
//					,working_buf_in
//					,*length
//				);
//				putsUSBUSART(output_buffer);
//			}

			// Save our new output pointer
			working_buf_out = tst_char;
			return (TRUE);
		}
		
		// Also check for running into input pointer
		if (tst_char == working_buf_in)
		{
//			if (mUSBUSARTIsTxTrfReady())
//			{
//				sprintf( 
//					output_buffer, 
//					(rom char*)"END:\r"
//				);
//				putsUSBUSART(output_buffer);
//			}

			// Treat this condition as - nope - we don't 
			// have a real packet
			break;
		}
	}

//	if (mUSBUSARTIsTxTrfReady())
//	{
//		sprintf( 
//			output_buffer, 
//			(rom char*)"END-end:\r"
///		);
//		putsUSBUSART(output_buffer);
//	}

	// We did not find our end, which is OK.
	*length = 0;
	input_buffer[0] = 0x00;
	return (FALSE);
}

// Look at the new packet, see what command it is, and 
// route it appropriately.
void parse_packet(unsigned char length)
{
	// Do preliminary sanity checks on packet
	// If length is only 1 byte, then it's a bad packet (need at least a letter and <CR>)
	// If length is 2, then need a <CR> at end
	// If length > 2, then comma must be second or third byte
	if (
		(1 == length)
		||
		(
			(length > 2)
			&&
			(input_buffer[1] != ',')
			&&
			(input_buffer[2] != ',')
		)
	)
	{
		g_error = ',';
		return;
	}

	// If last character is not <CR>, this isn't a good packet.
	if (input_buffer[length - 1] != 13)
	{
		g_error = '!';
		return;
	}
	
	// Check for valid command type as first byte
	switch (input_buffer[0])
	{		
		case 'R':
		{
			switch (input_buffer[1])
			{
				case 'X':
				{
					parse_RX_packet();		// For receiving serial
					break;
				}
				
				default:
				{
					// Reset command (resets everything to power-on state)
					parse_R_packet();
					break;
				}
			}
			break;
		}
		
		case 'C':
		{
			switch (input_buffer[1])
			{
				case ',':
				{
					// Configure command (configure ports for Input or Ouptut)
					parse_C_packet();
					break;
				}
				
				case 'X':
				{
					parse_CX_packet();		// For configuring serial port
					break;
				}
				
				default:
				{
					// Send back 'unknown command' error
					g_error = '?';
					return;	
					break;
				}
			}
			break;
		}

		// Output command (tell the ports to output something)
		case 'O':
		{
			parse_O_packet();
			break;
		}

		// Input command (return the current status of the ports)
		case 'I':
		{
			parse_I_packet();
			break;
		}

		// Version command
		case 'V':
		{
			parse_V_packet();
			break;
		}

		// Analog command
		case 'A':
		{
			parse_A_packet();
			break;
		}

		case 'T':
		{
			switch (input_buffer[1])
			{
				case ',':
				{
					parse_T_packet();		// For timed I/O
					break;
				}
				
				case 'X':
				{
					parse_TX_packet();		// For transmitting serial
					break;
				}
				
				default:
				{
					// Send back 'unknown command' error
					g_error = '?';
					return;	
					break;
				}
			}
			break;
		}
			
		case 'P':
		{
			switch (input_buffer[1])
			{
				case 'I':
				{
					parse_PI_packet();		// PI for reading a single pin
					break;
				}
				
				case 'O':
				{
					parse_PO_packet();		// PO for setting a single pin
					break;
				}
				
				case 'D':
				{
					parse_PD_packet();		// PD for setting a pin's direction
					break;
				}
				
				default:
				{
					// Send back 'unknown command' error
					g_error = '?';
					return;	
					break;
				}
			}
			break;
		}

		case 'M':
		{
			switch (input_buffer[1])
			{
				case 'R':
				{
					parse_MR_packet();		// MR for Memory Read
					break;
				}
				
				case 'W':
				{
					parse_MW_packet();		// MW for Memory Write
					break;
				}
				
				default:
				{
					// Send back 'unknown command' error
					g_error = '?';
					return;	
					break;
				}
			}
			break;
		}
		
		default:
		{
			// Send back 'unknown command' error
			g_error = '?';
			return;	
			
			break;
		}
	}
}

// Return all I/Os to their default power-on values
void parse_R_packet(void)
{
	UserInit();

	if (mUSBUSARTIsTxTrfReady())
	{
		output_buffer[0] = 'O';
		output_buffer[1] = 'K';
		output_buffer[2] = '\n';
		output_buffer[3] = '\r';
		output_buffer[4] = 0x00;
		putsUSBUSART(output_buffer);
	}	
}

// "T" Packet
// Causes PIC to sample digital inputs at a regular interval and send
// I packets back at that interval.
// Send T,0,0<CR> to stop digital
// FORMAT: T,<TIME_BETWEEN_UPDATES_IN_MS>,<MODE><CR>
// <MODE> is 0 for digital (I packets) and 1 for analog (A packets)
// EXAMPLE: "T,4000,0<CR>" to send an I packet back every 4 seconds.
// EXAMPLE: "T,2000,1<CR>" to send an A packet back every 2 seconds.
void parse_T_packet(void)
{
	unsigned char ptr = 1;
	unsigned char mode = 0;
	
	// Extract each of the value.
	time_between_updates = extract_number (&ptr, kUINT);
	mode = extract_number (&ptr, kUCHAR);

	if (0 != g_error)
	{
		return;
	}

	// Check to make sure our packet was the right length.
	if (input_buffer[ptr] != 13 && (0 == g_error))
	{
		g_error = '#';
		return;
	}

	// Now start up the timer at the right rate or shut 
	// it down.
	if (0 == mode)
	{
		if (0 == time_between_updates)
		{
			// Turn off sending of I packets.
			ISR_D_RepeatRate = 0;
		}
		else
		{
			// Limit our time between updates value to >= 10
			if (time_between_updates < 10)
			{
				time_between_updates = 10;
			}
			
			T2CONbits.TMR2ON = 1;    
		
			// Eventually gaurd this section from interrupts
			ISR_D_RepeatRate = time_between_updates;
		}
	}	
	else
	{
		if (0 == time_between_updates)
		{
			// Turn off sending of A packets.
			ISR_A_RepeatRate = 0;
		}
		else
		{
			// Limit our time between updates value to >= 10
			if (time_between_updates < 10)
			{
				time_between_updates = 10;
			}
		
			T2CONbits.TMR2ON = 1;    
		
			// Eventually gaurd this section from interrupts
			ISR_A_RepeatRate = time_between_updates;
		}
	}

	if (mUSBUSARTIsTxTrfReady())
	{
		output_buffer[0] = 'O';
		output_buffer[1] = 'K';
		output_buffer[2] = '\n';
		output_buffer[3] = '\r';
		output_buffer[4] = 0x00;
		putsUSBUSART(output_buffer);
	}	
}



// FORMAT: C,<portA_IO>,<portB_IO>,<portC_IO>,<analog_config><CR>
// EXAMPLE: "C,255,0,4,0<CR>"
// <portX_IO> is the byte sent to the Data Direction (DDR) regsiter for
// each port. A 1 in a bit location means input, a 0 means output.
// <analog_config> is a value between 0 and 12. It tells the UBW
// how many analog inputs to enable. If a zero is sent for this 
// parameter, all analog inputs are disabled.
// For the other values, see the following chart to know what pins are 
// used for what:
// 
// Note that in the following chart, PortE is references. This port
// only exists on the 40 and 44 pin versions of the UBW. For the 
// 28 pin versions of the UBW, all PortE based analog pins will return
// zero.
//
// <analog_config>	Analog Inputs Enabled	Pins Used For Analog Inputs
// ---------------	---------------------	-------------------------------
//	0				<none>					<none>
//	1				AN0						A0
//	2				AN0,AN1					A0,A1	
//	3				AN0,AN1,AN2				A0,A1,A2	
//	4				AN0,AN1,AN2,AN3			A0,A1,A2,A3	
//	5				AN0,AN1,AN2,AN3,AN4		A0,A1,A2,A3,A5		
//	6				AN0,AN1,AN2,AN3,AN4,	A0,A1,A2,A3,A5,E0
//						AN5						
//	7				AN0,AN1,AN2,AN3,AN4,	A0,A1,A2,A3,A5,E0,E1
//						AN5,AN6						
//	8				AN0,AN1,AN2,AN3,AN4,	A0,A1,A2,A3,A5,E0,E1,E2
//						AN5,AN6,AN7						
//	9				AN0,AN1,AN2,AN3,AN4,	A0,A1,A2,A3,A5,E0,E1,E2,B2
//						AN5,AN6,AN7,AN8						
//	10				AN0,AN1,AN2,AN3,AN4,	A0,A1,A2,A3,A5,E0,E1,E2,B2,B3
//						AN5,AN6,AN7,AN8,
//						AN9						
//	11				AN0,AN1,AN2,AN3,AN4,	A0,A1,A2,A3,A5,E0,E1,E2,B2,B3,B1
//						AN5,AN6,AN7,AN8,
//						AB9,AN10						
//	12				AN0,AN1,AN2,AN3,AN4,	A0,A1,A2,A3,A5,E0,E1,E2,B2,B3,B1,B4
//						AN5,AN6,AN7,AN8,
//						AN9,AN10,AN11
// NOTE: it is up to the user to tell the proper port direction bits to be
// inputs for the analog channels they wish to use.
void parse_C_packet(void)
{
	unsigned char PA, PB, PC, AA;
	unsigned char ptr = 1;

	// Extract each of the four values.
	PA = extract_number (&ptr, kUCHAR);
	PB = extract_number (&ptr, kUCHAR);
	PC = extract_number (&ptr, kUCHAR);
	AA = extract_number (&ptr, kUCHAR);

	if (0 != g_error)
	{
		return;
	}

	// Check to make sure our packet was the right length.
	if (input_buffer[ptr] != 13 && (0 == g_error))
	{
		g_error = '#';
		return;
	}

	if (mUSBUSARTIsTxTrfReady())
	{
		output_buffer[0] = 'O';
		output_buffer[1] = 'K';
		output_buffer[2] = '\n';
		output_buffer[3] = '\r';
		output_buffer[4] = 0x00;
		putsUSBUSART(output_buffer);
	}	

	// Now write those values to the data direction registers.
	TRISA = PA;
	TRISB = PB;
	TRISC = PC;

	// Handle the analog value.
	// Maximum value of 12.
	if (AA > 12)
	{
		AA = 12;
	}

	// If we are turning off Analog inputs
	if (0 == AA)
	{
		// Turn all analog inputs into digital inputs
		ADCON1 = 0x0F;
		// Turn off the ADC
		ADCON0bits.ADON = 0;
		// Turn off our own idea of how many analog channels to convert
		AnalogEnable = 0;
	}
	else
	{
		// Some protection from ISR
		AnalogEnable = 0;

		// We're turning some on.
		// Start by selecting channel zero		
		ADCON0 = 0;

		// Then enabling the proper number of channels
		ADCON1 = 15 - AA;

		// Set up ADCON2 options
		// A/D Result right justified
		// Acq time = 20 Tad (?)
		// Tad = Fosc/64
		ADCON2 = 0b10111110;

		// Turn on the ADC
		ADCON0bits.ADON = 1;

		// Tell ourselves how many channels to convert, and turn on ISR conversions
		AnalogEnable = AA;

		T2CONbits.TMR2ON = 1;    

	}

}

// Outputs values to the ports pins that are set up as outputs.
// Example "O,121,224,002<CR>"
void parse_O_packet(void)
{
	unsigned char PA, PB, PC;
	unsigned char ptr = 1;

	// Extract each of the four values.
	PA = extract_number (&ptr, kUCHAR);
	PB = extract_number (&ptr, kUCHAR);
	PC = extract_number (&ptr, kUCHAR);

	if (0 != g_error)
	{
		return;
	}

	// Check to make sure our packet was the right length.
	if (input_buffer[ptr] != 13 && (0 == g_error))
	{
		g_error = '#';
		return;
	}

	if (0 == g_error)
	{
		// Now write those values to the data port registers.
		LATA = PA;
		LATB = PB;
		LATC = PC;
	}

	if (mUSBUSARTIsTxTrfReady())
	{
		output_buffer[0] = 'O';
		output_buffer[1] = 'K';
		output_buffer[2] = '\n';
		output_buffer[3] = '\r';
		output_buffer[4] = 0x00;
		putsUSBUSART(output_buffer);
	}	
}

// Read in the three I/O ports (A,B,C) and create
// a packet to send back with all of values.
// Example: "I,143,221,010<CR>"
// Remember that on UBW 28 pin boards, we only have
// Port A bits 0 through 5
// Port B bits 0 through 7
// Port C bits 0,1,2 and 7,8
// And that Port C bits 0,1,2 are used for
// 		User1 LED, User2 LED and Program switch respectively.
// The rest will be read in as zeros.
void parse_I_packet(void)
{
	if (mUSBUSARTIsTxTrfReady())
	{
		sprintf (
			(char *)output_buffer, 
			(rom char*)"I,%03i,%03i,%03i\n\r", 
			PORTA,
			PORTB,
			PORTC
		);
		putsUSBUSART(output_buffer);
	}	
}

// All we do here is just print out our version number
void parse_V_packet(void)
{
	if (mUSBUSARTIsTxTrfReady())
	{
		putrsUSBUSART(g_version);
	}
}

// A is for read Analog inputs
// Just print out the last analog values for each of the
// enabled channels. The number of value returned in the
// A packet depend upon the number of analog inputs enabled.
// The user can enabled any number of analog inputs between 
// 0 and 12. (none enabled, through all 12 analog inputs enabled).
// Returned packet will look like "A,0,0,0,0,0,0<CR>" if
// six analog inputs are enabled but they are all
// grounded. Note that each one is a 10 bit
// value, where 0 means the intput was at ground, and
// 1024 means it was at +5 V. (Or whatever the USB +5 
// pin is at.) 
void parse_A_packet(void)
{
	char pos = 0;
	char channel = 0;

	if (mUSBUSARTIsTxTrfReady())
	{
		// Put the beginning of the packet in place
		output_buffer[pos] = 'A';
		pos++;

		// Now add each analog value
		for (channel = 0; channel < AnalogEnable; channel++)
		{
			sprintf(
				 (char *)&output_buffer[pos]
				,(rom char *)",%04u" 
				,ISR_A_FIFO[channel][0]
			);
			pos += 5;
		}

		// For debugging
//		sprintf(
//			 (char *)&output_buffer[pos]
//			,(rom char *)",%04u" 
//			,IntCnt
//		);
//		pos += 5;

		// Add \n\r and terminating zero.
		output_buffer[pos] = 10;
		pos++;
		output_buffer[pos] = 13;
		pos++;
		output_buffer[pos] = 0x00;

		putsUSBUSART(output_buffer);
	}	
}

// MW is for Memory Write
// "MW,<location>,<value><CR>"
// <location> is a decimal value between 0 and 4096 indicating the RAM address to write to 
// <value> is a decimal value between 0 and 255 that is the value to write
void parse_MW_packet(void)
{
	unsigned int location;
	unsigned char value;
	unsigned char ptr = 2;

	location = extract_number (&ptr, kUINT);
	value = extract_number (&ptr, kUCHAR);

	if (0 != g_error)
	{
		return;
	}

	// Check to make sure our packet was the right length.
	if (input_buffer[ptr] != 13 && (0 == g_error))
	{
		g_error = '#';
		return;
	}

	// Limit check the address and write the byte in
	if (location < 4096)
	{
		*((unsigned char *)location) = value;
	}

	if (mUSBUSARTIsTxTrfReady())
	{
		output_buffer[0] = 'O';
		output_buffer[1] = 'K';
		output_buffer[2] = '\n';
		output_buffer[3] = '\r';
		output_buffer[4] = 0x00;
		putsUSBUSART(output_buffer);
	}	
}


// MR is for Memory Read
// "MW,<location><CR>"
// <location> is a decimal value between 0 and 4096 indicating the RAM address to read from 
// The UBW will then send a "MR,<value><CR>" packet back to the PC
// where <value> is the byte value read from the address
void parse_MR_packet(void)
{
	unsigned int location;
	unsigned char value;
	unsigned char ptr = 2;

	location = extract_number (&ptr, kUINT);

	if (0 != g_error)
	{
		return;
	}

	// Check to make sure our packet was the right length.
	if (input_buffer[ptr] != 13 && (0 == g_error))
	{
		g_error = '#';
		return;
	}

	// Limit check the address and write the byte in
	if (location < 4096)
	{
		value = *((unsigned char *)location);
	}

	// Now send back the MR packet
	if (mUSBUSARTIsTxTrfReady())
	{
		sprintf(
			 (char *)output_buffer
			,(rom char *)"MR,%03u\n\r" 
			,value
		);

		putsUSBUSART(output_buffer);
	}	
}

// PD is for Pin Direction
// "PD,<port>,<pin>,<direction><CR>"
// <port> is "A", "B", "C" and indicates the port
// <pin> is a number between 0 and 7 and indicates which pin to change direction on
// <direction> is "1" for input, "0" for output
void parse_PD_packet(void)
{
	unsigned char port;
	unsigned char pin;
	unsigned char direction;
	unsigned char ptr = 2;

	port = extract_number (&ptr, kASCII_CHAR);
	pin = extract_number (&ptr, kUCHAR);
	direction = extract_number (&ptr, kUCHAR);
	
	if (0 != g_error)
	{
		return;
	}

	// Check to make sure our packet was the right length.
	if (input_buffer[ptr] != 13 && (0 == g_error))
	{
		g_error = '#';
		return;
	}

	// Limit check the parameters
	if (direction > 1)
	{
		g_error = '*';
		return;
	}
	if (pin > 7)
	{
		g_error = '*';
		return;
	}
	if ('A' == port)
	{
		if (0 == direction)
		{
			bitclr (DDRA, pin);  	
		}
		else
		{
			bitset (DDRA, pin);  	
		}
	}
	else if ('B' == port)
	{
		if (0 == direction)
		{
			bitclr (DDRB, pin);  	
		}
		else
		{
			bitset (DDRB, pin);  	
		}		
	}
	else if ('C' == port)
	{
		if (0 == direction)
		{
			bitclr (DDRC, pin);  	
		}
		else
		{
			bitset (DDRC, pin);  	
		}		
	}
	else
	{
		g_error = '*';
		return;	
	}

	if (mUSBUSARTIsTxTrfReady())
	{
		output_buffer[0] = 'O';
		output_buffer[1] = 'K';
		output_buffer[2] = '\n';
		output_buffer[3] = '\r';
		output_buffer[4] = 0x00;
		putsUSBUSART(output_buffer);
	}
}

// PI is for Pin Input
// "PI,<port>,<pin><CR>"
// <port> is "A", "B", "C" and indicates the port
// <pin> is a number between 0 and 7 and indicates which pin to read
// The command returns a "PI,<value><CR>" packet,
// where <value> is the value (0 or 1 for digital, 0 to 1024 for Analog)
// value for that pin.
void parse_PI_packet(void)
{
	unsigned char port;
	unsigned char pin;
	unsigned char ptr = 2;
	unsigned char value = 0;

	port = extract_number (&ptr, kASCII_CHAR);
	pin = extract_number (&ptr, kUCHAR);

	if (0 != g_error)
	{
		return;
	}

	// Check to make sure our packet was the right length.
	if (input_buffer[ptr] != 13 && (0 == g_error))
	{
		g_error = '#';
		return;
	}

	// Limit check the parameters
	if (pin > 7)
	{
		g_error = '*';
		return;
	}

	// Then test the bit in question based upon port
	if ('A' == port)
	{
		value = bittst (PORTA, pin);  	
	}
	else if ('B' == port)
	{
		value = bittst (PORTB, pin);  	
	}
	else if ('C' == port)
	{
		value = bittst (PORTC, pin);  	
	}
	else
	{
		g_error = '*';
		return;	
	}

	// Now send back our response
	if (mUSBUSARTIsTxTrfReady())
	{
		sprintf(
			 (char *)output_buffer
			,(rom char *)"PI,%1u\n\r" 
			,value
		);

		putsUSBUSART(output_buffer);
	}	

}

// PO is for Pin Output
// "PO,<port>,<pin>,<value><CR>"
// <port> is "A", "B", "C" and indicates the port
// <pin> is a number between 0 and 7 and indicates which pin to write out the value to
// <value> is "1" or "0" and indicates the state to change the pin to
void parse_PO_packet(void)
{
	unsigned char port;
	unsigned char pin;
	unsigned char value;
	unsigned char ptr = 2;

	port = extract_number (&ptr, kASCII_CHAR);
	pin = extract_number (&ptr, kUCHAR);
	value = extract_number (&ptr, kUCHAR);

	if (0 != g_error)
	{
		return;
	}

	// Check to make sure our packet was the right length.
	if (input_buffer[ptr] != 13 && (0 == g_error))
	{
		g_error = '#';
		return;
	}

	// Limit check the parameters
	if (value > 1)
	{
		g_error = '*';
		return;
	}
	if (pin > 7)
	{
		g_error = '*';
		return;
	}
	if ('A' == port)
	{
		if (0 == value)
		{
			bitclr (LATA, pin);  	
		}
		else
		{
			bitset (LATA, pin);  	
		}
	}
	else if ('B' == port)
	{
		if (0 == value)
		{
			bitclr (LATB, pin);  	
		}
		else
		{
			bitset (LATB, pin);  	
		}		
	}
	else if ('C' == port)
	{
		if (0 == value)
		{
			bitclr (LATC, pin);  	
		}
		else
		{
			bitset (LATC, pin);  	
		}		
	}
	else
	{
		g_error = '*';
		return;	
	}

	if (mUSBUSARTIsTxTrfReady())
	{
		output_buffer[0] = 'O';
		output_buffer[1] = 'K';
		output_buffer[2] = '\n';
		output_buffer[3] = '\r';
		output_buffer[4] = 0x00;
		putsUSBUSART(output_buffer);
	}	
}

// TX is for Serial Transmit
// "TX,<data_length>,<variable_length_data><CR>"
// <data_length> is a count of the number of bytes in the <variable_length_data> field.
// It must never be larger than the number of bytes that are currently free in the
// software TX buffer or some data will get lost.
// <variable_length_data> are the bytes that you want the UBW to send. It will store them
// in its software TX buffer until there is time to send them out the TX pin.
// If you send in "0" for a <data_length" (and thus nothing for <variable_length_data>
// then the UBW will send back a "TX,<free_buffer_space><CR>" packet,
// where <free_buffer_space> is the number of bytes currently available in the 
// software TX buffer.
void parse_TX_packet(void)
{
}

// RX is for Serial Receive
// "RX,<length_request><CR>"
// <length_request> is the maximum number of characters that you want the UBW to send
// back to you in the RX packet. If you use "0" for <length_request> then the UBW
// will just send you the current number of bytes in it's RX buffer, and if
// there have been any buffer overruns since the last time a <length_request> of 
// "0" was received by the UBW.
// This command will send back a "RX,<length>,<variable_length_data><CR>"
// or "RX,<buffer_fullness>,<status><CR>" packet depending upon if you send
// "0" or something else for <length_request>
// <length> in the returning RX packet is a count of the number of bytes
// in the <variable_length_data> field. It will never be more than the
// <length_request> you sent in.
// <variable_length_data> is the data (in raw form - byte for byte what was received - 
// i.e. not translated in any way, into ASCII values or anything else) that the UBW
// received. This may include <CR>s and NULLs among any other bytes, so make sure
// your PC application treates the RX packet coming back from the UBW in a speical way
// so as not to screw up normal packet processing if any special caracters are received.
// <buffer_fullness> is a valule between 0 and MAX_SERIAL_RX_BUFFER_SIZE that records
// the total number of bytes, at that point in time, that the UBW is holding, waiting
// to pass on to the PC.
// <status> has several bits. 
//	Bit 0 = Software RX Buffer Overrun (1 means software RX buffer (on RX pin)
//		has been overrun and data has been lost) This will happen if you don't
//		read the data out of the UWB often enough and the data is coming in too fast.
//	Bit 1 = Software TX Buffer Overrun (1 means software TX buffer (on TX pin)
//		as been overrun and data hs been lost. This will happen if you send too much
//		data to the UBW and you have the serial port set to a low baud rate.
void parse_RX_packet(void)
{
}

// CX is for setting up serial port parameters
// TBD
void parse_CX_packet(void)
{
	if (mUSBUSARTIsTxTrfReady())
	{
		output_buffer[0] = 'O';
		output_buffer[1] = 'K';
		output_buffer[2] = '\n';
		output_buffer[3] = '\r';
		output_buffer[4] = 0x00;
		putsUSBUSART(output_buffer);
	}	
}

// RC is for outputting RC servo pulses on a pin
// "RC,<port>,<pin>,<value><CR>"
// <port> is "A", "B", "C" and indicates the port
// <pin> is a number between 0 and 7 and indicates which pin to output the new value on
// <value> is a decimal number between 0 and 1024.
// If <value> is "0" then the RC output on that pin is disabled.
// Otherwise <value> = 1 means 1ms pulse, <value> = 1024 means 2ms pulse,
// any value inbetween means proportional pulse values between those two
// Note: The pin used for RC output must be set as an output, or not much will happen.
// The RC command will continue to send out pulses at the last set value on 
// each pin that has RC output with a repition rate of 1 pulse every 20ms.
// If you have RC output enabled on a pin, outputting a digital value to that pin
// will be overwritten the next time the RC pulses. Make sure to turn off the RC
// output if you want to use the pin for something else.
void parse_RC_packet(void)
{
	if (mUSBUSARTIsTxTrfReady())
	{
		output_buffer[0] = 'O';
		output_buffer[1] = 'K';
		output_buffer[2] = '\n';
		output_buffer[3] = '\r';
		output_buffer[4] = 0x00;
		putsUSBUSART(output_buffer);
	}	
}

// Look at the string pointed to by ptr
// There should be a comma where ptr points to upon entry.
// If not, throw a comma error.
// If so, then look for up to three bytes after the
// comma for numbers, and put them all into one
// byte (0-255). If the number is greater than 255, then
// thow a range error.
// Advance the pointer to the byte after the last number
// and return.
signed short long extract_number(unsigned char * ptr, unsigned char type)
{
	signed short long acc;
	unsigned char neg = FALSE;
	
	// If we already have an error, don't do anything here.
	if (0 != g_error)
	{
		return (0);
	}

	// Check for comma where ptr points
	if (input_buffer[*ptr] != ',')
	{
		g_error = '^';
		return(0);
	}
	// Incriment the pointer past the comma
	(*ptr)++;

	// Now check for a sign character
	if (('-' == input_buffer[*ptr]) && (kASCII_CHAR != type))
	{
		// It's an error if we see a negative sign on an unsigned value
		if (
			(kUCHAR == type)
			||
			(kUINT == type)
		)
		{
			g_error = '|';
			return (0);
		}
		else
		{
			neg = TRUE;
			(*ptr)++;
		}
	}

	// If we need to get a digit, go do that
	if (kASCII_CHAR != type)
	{
		extract_digit(ptr, &acc, 5);
	}
	else
	{
		// Otherwise just copy the byte
		acc = input_buffer[*ptr];
		(*ptr)++;
	}

	// Handle the negative sign
	if (neg)
	{
		acc = -acc;
	}

	// Range check the new value
	if (
		(
			kCHAR == type
			&&
			(
				(acc > 127)
				||
				(acc < -128)
			)
		)
		||
		(
			kUCHAR == type
			&&
			(
				(acc > 255)
				||
				(acc < 0)
			)
		)
		||
		(
			kINT == type
			&&
			(
				(acc > 32767)
				||
				(acc < -32768)
			)
		)
		||
		(
			kUINT == type
			&&
			(
				(acc > 65535)
				||
				(acc < 0)
			)
		)
	)
	{
		if (0 == g_error)
		{
			g_error = '~';
			return (0);
		}
	}

	return(acc);	
}

// Loop 'digits' number of times, looking at the
// byte in input_buffer index *ptr, and if it is
// a digit, adding it to acc. Take care of 
// powers of ten as well. If you hit a non-numerical
// char, then return FALSE, otherwise return TRUE.
// Store result as you go in *acc.
signed char extract_digit(unsigned char * ptr, signed short long * acc,	unsigned char digits)
{
	unsigned char val;
	unsigned char digit_cnt;
	
	*acc = 0;

	for (digit_cnt = 0; digit_cnt < digits; digit_cnt++)
	{
		val = input_buffer[*ptr];
		if ((val >= 48) && (val <= 57))
		{
			*acc = (*acc * 10) + (val - 48);
			(*ptr)++;
		}
		else
		{
			return (FALSE);
		}
	}
	return (TRUE);
}

// For debugging, this command will spit out a bunch of values.
void print_status(void)
{
	if (g_error == 0)
	{	
		if (mUSBUSARTIsTxTrfReady())
		{
			sprintf( 
				 (char *)output_buffer
				,(rom char*)"Status=%i\n\r"
				,ISR_D_FIFO_length
			);
			putsUSBUSART(output_buffer);
		}
	}
}
/******************************************************************************
 * Function:        void BlinkUSBStatus(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        BlinkUSBStatus turns on and off LEDs corresponding to
 *                  the USB device state.
 *
 * Note:            mLED macros can be found in io_cfg.h
 *                  usb_device_state is declared in usbmmap.c and is modified
 *                  in usbdrv.c, usbctrltrf.c, and usb9.c
 *****************************************************************************/
void BlinkUSBStatus(void)
{
    static word led_count = 0;
    
    if (0 == led_count)
	{
		led_count = 50000U;
	}
    led_count--;

    if (
		usb_device_state == DETACHED_STATE
       	||
       	1 == UCONbits.SUSPND
    )
    {
		mLED_1_Off();
    }
    else if (
		usb_device_state == ATTACHED_STATE
		||
		usb_device_state == POWERED_STATE		
		||
		usb_device_state == DEFAULT_STATE
		||
		usb_device_state == ADDRESS_STATE
	)
    {
        mLED_1_On();
    }
    else if (usb_device_state == CONFIGURED_STATE)
    {
        if (0 == led_count)
        {
			mLED_1_Toggle();
        }
    }
}

BOOL SwitchIsPressed(void)
{
	if( 0 == UserSW)                   	// If pressed
	{
	    return (TRUE);                	// Was pressed
	}
	else
	{
		return (FALSE);					// Was not pressed
	}
}

/** EOF user.c ***************************************************************/
